#pragma once
#include <math.h>
#include "MatrixAlgebra.h"

// Bezier surface   wikipedia :
//   http://en.wikipedia.org/wiki/B%C3%A9zier_surface


struct sPoint {
	double x, y, z;

	sPoint (double x_=0, double y_=0, double z_=0) : x(x_), y(y_), z(z_) { }
};

inline double factorial (int n)
{
	double s = 1;
	for (int i=2; i<=n; i++) s *= i;
	return s;
}

inline double BernsteinPolynomial(int n, int i, double u)
{
	return factorial(n)/factorial(i)/factorial(n-i)*pow(u,i)*pow(1-u,n-i);
}


struct sBezierSurface {
	double xmin, xmax;
	double ymin, ymax;
	double cp[5][5]; // Bezier Surface 5x5 control points

	void BezierSurfaceFitting (sPoint *points, int no_points)
	{
		dMatrix Q(no_points, 1);
		dMatrix B(no_points, 5*5);
		dMatrix P(5*5,       1);

		xmin = ymin = +1.e10;
		xmax = ymax = -1.e10;

		// x, y  ִ, ּ  ã´.
		// ִ, ּ   x, y  0 ~ 1  ȭ ϱ Ѵ.
		sPoint *p = points;
		for (int c=0; c<no_points; c++) {
			xmin = std::min (xmin, p->x);
			xmax = std::max (xmax, p->x);
			ymin = std::min (ymin, p->y);
			ymax = std::max (ymax, p->y);
			p++;
		}

		p = points;
		for (int c=0; c<no_points; c++) {
			Q(c,0) = p->z;
			
			double u = (double)(p->x - xmin)/(xmax - xmin);
			double v = (double)(p->y - ymin)/(ymax - ymin);

			int k = 0;
			for (int i=0; i<=4; i++) {
				for (int j=0; j<=4; j++) {
					B(c,k) = BernsteinPolynomial(4,i,u)*BernsteinPolynomial(4,j,v);
					k++;
				}
			}
			p++;
		}

		// Pseudo-inverse ظ Ѵ.
		P = !(~B*B)*~B*Q;

		int k = 0;
		for (int i=0; i<=4; i++) {
			for (int j=0; j<=4; j++) {
				cp[i][j] = P(k,0);
				k++;
			}
		}
	}

	double BezierSurfaceZ (double x, double y)
	{
		double u = (double)(x - xmin)/(xmax - xmin);
		double v = (double)(y - ymin)/(ymax - ymin);

		// ȭ x, y κ Bezier Surface z Ѵ.
		double z = 0;
		for (int i=0; i<=4; i++) {
			for (int j=0; j<=4; j++) {
				z += BernsteinPolynomial(4,i,u)*BernsteinPolynomial(4,j,v)*cp[i][j];
			}
		}
		return z;
	}

	inline double length (sPoint &pt)
	{
		//  ptκΤ Bezier Surface Z-   ̸ Ѵ.
		return fabs(BezierSurfaceZ(pt.x,pt.y) - pt.z);
	}
};

extern double bezier_surface_fitting (sPoint *data, int no_data, sBezierSurface &model, double distance_threshold);
