// LidarData.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdio.h>
#include <assert.h>
#include <opencv2/opencv.hpp>
#include "BezierSurfaceFit.h"

#pragma pack(1)

struct sElement {
	short			x, y, z;	//  x,y,z ġ, : cm
	unsigned short	i, e;		// ( )
	unsigned short	h;			// horizontal angle,	ȯ: h/100,				: deg
	unsigned short	v;			// vertical angle,		ȯ: -30.666 + v*1.333,	: deg
	unsigned short	r;			// range points,		ȯ: r/5,				: cm
};

/*
	horizontal angel h   : 
	0: -30.67
	1: -29.33
	2: -28.00
	3: -26.66
	4: -25.33
	...
*/

bool IsNormalVectorVertical (sElement &xl, sElement &xr, sElement &yl, sElement&yr)
{
	// Cross product  :
	// (x3, y3, z3) = Cross (xl-xr, yl-yr);

	double x1 = xl.x - xr.x;
	double y1 = xl.y - xr.y;
	double z1 = xl.z - xr.z;

	double x2 = yl.x - yr.x;
	double y2 = yl.y - yr.y;
	double z2 = yl.z - yr.z;

	double x3 = y1*z2 - z1*y2;
	double y3 = z1*x2 - x1*z2;
	double z3 = x1*y2 - y1*x2;

	// x, y ⿡  z ũⰡ 5 ̻ ū 
	return fabs(z3) > 5*sqrt(x3*x3 + y3*y3);
}

#define R_MIN		 1000		// 2m
#define R_MAX		 10000		// 20m
#define X_MIN		-1000		// -10m
#define X_MAX		 1000		// 10m
#define Y_MIN		-1000		// -10m
#define Y_MAX		 1000		// 10m

// Velodyne HDL-32E  ͸ о  ޸
sElement pcd[3000][32];
bool     npcd[3000][32];	//  ĺ  true  

// LidarData.cpp : Defines the entry point for the console application.
//


int main()
{
	cvNamedWindow ("Image", CV_WINDOW_AUTOSIZE);

	FILE *fp = fopen ("Lidar.ldt", "rb");
	if (!fp) return -1;

	unsigned long header[60];
	fread (header, sizeof(unsigned long), 19, fp);

	int pcd_count = 0;
	while (!feof(fp)) {
		fread (header, sizeof(unsigned long), 56, fp);

		int no_frame_data = header[55];

		for (int i=0; i<no_frame_data/32; i++) {
			sElement frame[32];
			fread (&frame, sizeof(sElement), 32, fp);

			//  ʹ vertical  0,16,1,17,2,18,...  Ǿ ִ.
			// ̸  ٲپ Ѵ.
			for (int j=0; j<16; j++) {
				pcd[i][j+0 ] = frame[j*2+0];
				pcd[i][j+16] = frame[j*2+1];
			}
		}

		for (int i=0; i<3000; i++) {
			for (int j=0; j<32; j++) {
				npcd[i][j] = false;
			}
		}

		// о   Ϳ   ĺ  Ѵ.
		//  ġ յ, ¿ κ  ͸ Ͽ  Ͱ     ĺ  ȴ.
		int point_count = 0;
		sPoint *points = new sPoint [no_frame_data];

		int dx = 5, dy = 1;
		for (int i=dx; i<no_frame_data/32-dx; i++) {
			for (int j=dy; j<32-dy; j++) {
				if (IsNormalVectorVertical (pcd[i+dx][j], pcd[i-dx][j], pcd[i][j+dy], pcd[i][j-dy])) {
					npcd[i][j] = true;
				}
			}
		}

		//  ĺ  ߿ յ¿ 4   ĺ    鿡 شϴ  .
		for (int i=dx; i<no_frame_data/32-dx; i++) {
			for (int j=dy; j<32-dy; j++) {
				sElement &pt = pcd[i][j];
				if (R_MIN <  pt.r && pt.r <  R_MAX &&
					X_MIN <= pt.x && pt.x <= X_MAX && 
					Y_MIN <= pt.y && pt.y <= Y_MAX) {
					if (rand()%10 == 0) {	//   ó ʰ 10  1  Ȯ ĺ Ѵ.
						if (npcd[i][j] && npcd[i-1][j] && npcd[i+1][j] && npcd[i][j-1] && npcd[i][j+1]) {
							points[point_count++] = sPoint (pt.x, pt.y, pt.z);
						}
					}
				}
			}
		}

		//   ã 鿡 شϴ κ Bezier Surface ǥϴ 5x5 Ʈ Ʈ Ѵ.
		// ⼭ RANSAC ˰  ʴ´. ֳϸ, 25 Ʈ Ʈ  Bezier Surface ã
		// ؼ 25 ̻  Ʈ ʿѵ,  Ʈ  RANSAC ˰ ȿ ϸ
		// Bezier Surface ϴµ ص  ǻ ð ҸǱ ̴.
		sBezierSurface model;
		double threshold = 7.5;	// cm
		double ret = bezier_surface_fitting (points, point_count, model, threshold);

		//  360 ȸϸ鼭 32üο  ͸ ׸  ̹ .
		// ̹  δ 10 ȮϿ ǥѴ.  32ü ͸ 320pixel ǥ
		CvSize size = cvSize(no_frame_data/32, 320);
		IplImage* img = cvCreateImage (size, IPL_DEPTH_8U, 1);

		cvSet (img, cvScalar(0));

		for (int i=dx; i<no_frame_data/32-dx; i++) {
			for (int j=dy; j<32-dy; j++) {
				sElement &pt = pcd[i][j];
				if (R_MIN <  pt.r && pt.r <  R_MAX &&
					X_MIN <= pt.x && pt.x <= X_MAX && 
					Y_MIN <= pt.y && pt.y <= Y_MAX) {

					// Ÿ rκ  :  ָ   Ӵ.
					int v = 255 - (int)(255*((double)pt.r/R_MAX));
					v = std::max (0, v);
					v = std::min (255, v);

					// horizontal  ̳   Ÿ(ȣ) (r)  0.016 ũٸ outlier .
					double outlier_threshold = 0.016;
					double v1 = (double)abs(pcd[i-1][j].r - pt.r)/pt.r;
					double v2 = (double)abs(pcd[i+1][j].r - pt.r)/pt.r;
					if (outlier_threshold < v1 && outlier_threshold < v2) continue;

					double l = model.length (sPoint(pt.x, pt.y, pt.z));
					if (l < threshold) continue;

					CvScalar pixel; 
					pixel.val[0] = v;

					for (int k=0; k<10; k++) {
						cvSet2D(img, 320-1-(j*10+k), i, pixel); 
					}
				}
			}
		}

		delete [] points;

		cvShowImage ("Image", img);

		cvWaitKey(1);

		cvReleaseImage(&img);

		pcd_count++;
		printf ("=== %d === \n", pcd_count);
	}
	fclose (fp);

	cvDestroyWindow ("Image");

	return 0;
}


